/*
 * COPYRIGHT NOTICE
 * Copyright (c) Alteon Networks, Inc. 1996 
 * All rights reserved
 * 
 */
/*
 * HISTORY
 * $Log: CMedia.h,v $
 * Revision 1.1.2.3  1998/12/22  03:14:09  tibor
 * 	named section changes
 * 	[1998/12/22  02:13:17  tibor]
 *
 * Revision 1.1.2.2  1998/06/19  07:55:28  tibor
 * 	moved, renamed CFlash to CMedia, bug fixes
 * 	[1998/06/19  07:51:00  tibor]
 * 
 * Revision 1.1.2.1  1998/06/02  04:57:30  tibor
 * 	new genflash routines
 * 	[1998/06/02  02:32:59  tibor]
 * 
 * $EndLog$
 */

#include "media_regions.h"
#include "flashImageDef.h"
#include "imageDef.h"
#include "CElf.h"

/* 
 * Media is defined by the following data structures in a manor similar to 
 * C++ classes, or more genericly objects.  Objects are created by the new_*
 * routines, destroyed via free_*.  Initialization is done in init_*.  
 */


/* CMedia error codes, also 0 returned from new_* is an error */
#define CMEDIA_E_MALLOC 	-1  /* malloc returned 0 */
#define CMEDIA_E_PARM 		-2  /* int/user parameter error */
#define CMEDIA_E_MAP 		-3  /* map entry error */
#define CMEDIA_E_COMPRESS       -4  /* compression error */
#define CMEDIA_E_REGION_FULL	-5  /* region set full */
#define CMEDIA_E_REGION_BEGIN   -6  /* seek to before region beginning */
#define CMEDIA_E_FILE_NOTFND	-10 /* file or directory not found */
#define CMEDIA_E_DISKFULL	-11 /* disk full on write */
#define CMEDIA_E_EOF            -12 /* read past end of file */
#define CMEDIA_E_IO_ERROR	-13 /* other file error */

/*
 * This structure holds the actual media region image.  Because access may
 * be by byte or word a union is used to save the trouble of always casting.
 */
typedef struct mediaRegionImage {
    int nBytesUsed;		    /* number of bytes used in image
				     * 0  - valid image, all null
				     * >0 - bytes in use
				     * -1 - null image (default r/o image) */
    U8 *pLoc;			    /* last read/write location in image */
    int nRegionSize;		    /* malloc'ed region size */
    
    union {
	U8  *pByte;                    
	U32 *pWord;
    } bw;
} mediaRegionImage_t;

/* 
 * This structure encapsulates all information about a physical region 
 * including what logical group  as a kernel image, boot block, config blocks.
 */
typedef struct mediaRegionObject { 
    const imageMapEntry_t *pMapEntry;	/* map entry defines region constants */
    imageDefinition_t *pDefObj;		/* media definition constants */
    int phyRegionNum;			/* physical region number 0-n */
    int regionSetNum;			/* region number (0-num_regions) in set */
    int bytesReserved;			/* number of reserved bytes */
    int bytesWritable;			/* max bytes writable (size - reserved) */
    int dummy;
    mediaRegionImage_t *pImage;		/* actual image */
} mediaRegionObject_t;

/*
 * This structure holds the actual kernel/boot image in (un)compressed
 * form. Because access may be by byte or word a union is used to save the
 * trouble of always casting.
 */
typedef struct imageBlock {
    U8 *pNextFree;		/* next available loc in image */
    int nBytesAlloc;		/* malloc size for bw */ 
    union {
	U8 *pByte;                    
	U32 *pWord;
    } bw;			/* malloc'd image storage */
} imageBlock_t;

/*
 * This structure is a layout of the actual media region set that defines a
 * kernel.  The length of Image is phyHeader.K_Len
 */
typedef struct imageKernelObject {
    flashKernelHeader_t *pPhyHeader;      /* physical kernel info header */
    flashKernelSectionInfo_t *pSecHeader; /* actual uncompressed section info */
    const imageMapEntry_t *pMap;	  /* Map info for err check/msgs */
    int dummy;
} imageKernelObject_t;

typedef struct imageBootObject {
    media_bootstrap_region_t   *pBootstrap;
    media_code_offset_region_t *pCodeOffsets;
    media_manufact_region_t    *pManufact;
    media_oem_region_t	       *pOem;
} imageBootObject_t;

/*
 * This structure is a "base class" object to contain "derived classes"
 */
typedef struct imageObject {
    int objType;			/* type of derived class */
    imageDefinition_t *pDefObj;		/* media definition constants */
    imageBlock_t *pImage;		/* Full linear image */
    union {
	imageKernelObject_t kernObj;	/* kernel object derived class data */ 
	imageBootObject_t bootObj;	/* boot object derived class data */
    } derived;
} imageObject_t;

typedef struct regionSet {
    const imageMapEntry_t *pMap;	/* map entry(s) that define set */
    mediaRegionObject_t **RegionSet;	/* array of regions that define set */
    int numMapEntries;			/* num of maps define region set */
    int numRegions;			/* number of regions in set */
    int curRegion;			/* current region being access for r/w */
    int numRegionsUsed;			/* max curRegion as a result of write */
    int dummy;
    imageObject_t *iObj;		/* linear image of set */
} regionSet_t;

/* 
 * This structure defines a configured(built) media image.
 */
typedef struct mediaObject {
    imageDefinition_t *pDefObj;		/* media definition constants */
    mediaRegionObject_t **RegionList;	/* list of all gen'ed media regions */
    int numRegionSets;			/* number of region sets in array */
    regionSet_t **SetList;		/* array of RegionSets */
} mediaObject_t; 

/* build format */
#define CMEDIA_BUILD_FMT_TIGON  1
#define CMEDIA_BUILD_FMT_GNE    2

/* for dump_* methods, type of region dump to perform */
#define	CMEDIA_DUMP_DUMPRAW	1	/* dump full region size */
#define	CMEDIA_DUMP_DUMPLINEAR	2   	/* Dump region writable data only */

/* for dump_* methods, type of write format */
#define	CMEDIA_DUMP_WRITEBIN	1	/* raw binary */
#define	CMEDIA_DUMP_WRITEGDB	2	/* gdb readable */
#define	CMEDIA_DUMP_WRITETTP	3	/* ttp readable */
#define	CMEDIA_DUMP_WRITESIM	4	/* sim model readable */
#define CMEDIA_DUMP_WRITEHDR	5	/* header file generation of image */
  
/* 
 * prototypes 
 *
 * most functions deal the media objects.  the *_regionSet() are helper functions
 * that build a logical image of a software entity and writes it as a set of 
 * region images.
 */
extern int init_CMedia(imageDefinition_t *pDef, char *pathBase, int verbose_level, int format_type);

/*
 * Public Methods
 *
 *  these are the only functions that should be done outside of the class
 */

/* mediaObject methods */
extern mediaObject_t *new_mediaObject(imageDefinition_t *pDef);
extern void free_mediaObject(mediaObject_t *pObj);
extern int init_mediaObject(mediaObject_t *pObj);
extern int build_mediaObject(mediaObject_t *pObj);
extern regionSet_t *getRegionSet_mediaObject(mediaObject_t *pObj, const char *RegionName);
extern int dump_mediaObject(mediaObject_t *pObj, FILE *pFile, int typeWrite);
extern void setHdr_MediaObject(mediaObject_t *pObj, char *psPrefix);

/* regionSet methods */
extern regionSet_t * new_regionSet(mediaRegionObject_t *RegionSet[]);
extern void free_regionSet(regionSet_t *pObj);
extern int build_regionSet(regionSet_t *pObj);
extern int dump_regionSet(regionSet_t *pObj, FILE *pFile, int typeDump, int typeWrite);
