/*
 * COPYRIGHT NOTICE
 * Copyright (c) Alteon Networks, Inc. 1996, 1997
 * All rights reserved
 *
 */
/*
 * HISTORY
 * $Log: CElf.c $
 * Revision sb_tibor/2 1999/03/24 04:11:27 tibor
 * 	cleaned up htonl references
 * 
 * Revision 1.1.2.4  1999/02/16  23:56:56  wayne
 * 	Move alt_def.h to top to suppress NBBY warning
 * 	[1999/02/16  23:56:34  wayne]
 *
 * Revision 1.1.2.3  1998/12/22  03:14:24  tibor
 * 	named section changes
 * 	[1998/12/22  02:13:18  tibor]
 * 
 * Revision 1.1.2.2  1998/09/18  18:34:51  mef
 * 	include byteorder.h from different are for cygwin32.
 * 	[1998/09/18  18:31:58  mef]
 * 
 * Revision 1.1.2.1  1998/06/02  04:57:46  tibor
 * 	libutils.a interface routines, elf file read object
 * 	[1998/06/02  02:08:59  tibor]
 * 
 * Revision 1.1.2.2  1998/05/08  01:31:01  tibor
 * 	copied from ts/utils
 * 	[1998/05/08  00:52:34  tibor]
 * 
 * Revision 1.1.2.1  1997/05/05  21:09:56  dbender
 * 	Initial ODE submission
 * 	[1997/05/05  21:08:57  dbender]
 * 
 * $EndLog$
 */
/***************************************************************************
 *
 *  Name :      CElf.c
 *
 *              Reads an ELF file into data structures in memory.
 *
 ***************************************************************************/
#include "alt_def.h"
#include <stdio.h>
#include "sys/types.h"
#include <bfd.h>
#include <stdlib.h>
#include <string.h>
#include <netinet/in.h>
#include "CElf.h"

static int parseSecName(elfSection_t *pSec, const char *name);
static void sortSecAddr(elfFileData_t *pObj);

/*
 * Method:	new_elfFileData
 *
 * Description:	constructor for Elf file read object
 *
 * Parameters:	file	- name of file to read
 *
 * Returns:	0	- sucess
 *		-1	- malloc error
 */
elfFileData_t *
new_elfFileData(char *FName)
{
    elfFileData_t *pObj;

    if ((pObj = (elfFileData_t *)calloc(1,sizeof(elfFileData_t)))) {
	if((pObj->fName = (char *)malloc(strlen(FName)+1))) {
	    strcpy(pObj->fName, FName);
	} else {
	    free(pObj);
	    pObj = NULL;
	}
    }

    return pObj;
}

/*
 * Method:	free_elfFileData
 *
 * Description:	destructor for Elf file read object
 *
 * Parameters:	obj	- elf object
 *
 * Returns:	nothing
 */
void
free_elfFileData(elfFileData_t *pObj)
{
    elfSection_t **pSecList;
    elfSection_t *pSec;
    
    if (pObj->fName) {
	free(pObj->fName);

	/* free up all section related storage */
	if ((pSecList = pObj->pSecList)) {
	    while (*pSecList) {
		pSec = *pSecList;
		if (pSec->name) {
		    free(pSec->name);
		}
		if (pSec->data) {
		    free(pSec->data);
		}
		free(pSec);
		pSecList++;
	    }
	    free(pObj->pSecList);
	}
    }

    free(pObj);
}

/*
 * Method:	read_elfFileData
 *
 * Description:	build elf file image using only the text, data, rodata,
 *		bss, sbss sections.
 *
 * Parameters:	obj	- elf read object
 *
 * Returns:	0	- sucess
 *		-1	- malloc error
 *		-10	- elf file not found
 *		-13	- some i/o error
 */
int
read_elfFileData(elfFileData_t *pObj)
{
    bfd *pBfd;
    asection *pBfdSec;
    elfSection_t **pSecList;
    elfSection_t *pSec;
    int nLen;
    int nType;
    int rc = 0;

    if ((pBfd = bfd_openr(pObj->fName, 0)) == NULL) {
	(void) fprintf(stderr, "Couldn't open file %s\n", pObj->fName);
	return CELF_E_FILE_NOTFND;
    }

    if (bfd_check_format(pBfd, bfd_object) != true) {
	(void) fprintf(stderr, "Invalid file format %s\n", pObj->fName);
	rc = CELF_E_INVALID_FMT;
	goto gExit;
    }

    /* save the start address */
    pObj->startAddr = (caddr_t)bfd_get_start_address(pBfd);

    /* get storage for section list, this is max size, will be adjusted later */
    if (!(pObj->pSecList =
	  (elfSection_t **)calloc(pBfd->section_count, sizeof(elfSection_t)))) {
	rc = CELF_E_MALLOC;
	goto gExit;
    }

    pSecList = pObj->pSecList;
    
    for (pBfdSec = pBfd->sections; pBfdSec != NULL && !rc;
						    pBfdSec = pBfdSec->next) {  
	nLen = 0;
	nType = SECTION_DEF_TYPE_UNKN;
	if (strncmp(pBfdSec->name, ".text", 5) == 0) {
	    nType = SECTION_DEF_TYPE_TEXT;
	    nLen = 5;
	} else if (strncmp(pBfdSec->name, ".data", 5) == 0) {
	    nType = SECTION_DEF_TYPE_DATA;
	    nLen = 5;
	} else if (strncmp(pBfdSec->name, ".rodata", 7) == 0) {
	    nType = SECTION_DEF_TYPE_RODATA;
	    nLen = 7;
	} else if (strncmp(pBfdSec->name, ".bss", 4) == 0) {
	    nType = SECTION_DEF_TYPE_BSS;
	    nLen = 4;
	} else if (strncmp(pBfdSec->name, ".sbss", 5) == 0) {
	    nType = SECTION_DEF_TYPE_SBSS;
	    nLen = 5;
	}

	/* only save valid section types */
	if (nType) {
	    if (!(pSec = (elfSection_t *)calloc(1, sizeof(elfSection_t)))) {
		rc = CELF_E_MALLOC;
		goto gExit;
	    }
	    
	    *pSecList++ = pSec;
	    pSec->nSecType = nType;
	    pObj->numSections++;
	    
	    if (!(rc = parseSecName(pSec, pBfdSec->name + nLen))) {
	    
		pSec->bfdflags = (unsigned long)pBfdSec->flags;
		pSec->addr = (caddr_t)pBfdSec->vma;
		if ((pSec->len = bfd_section_size (pBfd, pBfdSec))
		    && (pSec->bfdflags & SEC_LOAD)) {
		    if (!(pSec->data =
			  (unsigned char *)malloc(bfd_section_size(pBfd, pBfdSec)))) {
			rc = CELF_E_MALLOC;
			goto gExit;
		    }

		    if (!bfd_get_section_contents(pBfd, pBfdSec, (PTR)pSec->data, 0,
						  bfd_section_size(pBfd, pBfdSec))) {
			rc = CELF_E_IO_ERROR;
			goto gExit;
		    }
		}
	    }
	}
	
    }

    /* sort the sections in acending order */
    sortSecAddr(pObj);

gExit:    
    bfd_close(pBfd);
    
    return rc;
}

/*
 * Function:	parseSecName
 *
 * Description:	parse the section name for extra information
 *		section naming should be
 *                 .[text|data|rodata|bss|sbss].name.[<flags>]
 *
 * Parameters:	obj	- section object
 *		name	- section name
 *
 * Returns:	0	- sucess
 *		-1	- malloc error
 */
static int
parseSecName(elfSection_t *pSec, const char *name)
{
    const char *pName;
    int nLen;

    /* no use doing anything if its just a standard section name
     * struct is assumed to be init'ed to null for name */
    if (!*name) {
	return 0;
    }
    
    /* first advance past inital dot */
    if (*name == '.') {
	name++;
    }

    /* actual parsing starts here */
    pName = name;

    while (*pName && *pName != '.') {
	pName++;
    }
    
    /* save away label of section name */
    if ((nLen = pName - name)) {
	if (!(pSec->name = (char *)malloc(nLen+1))) {
	    return CELF_E_MALLOC;
	}
	strncpy(pSec->name, name, nLen);
	*(pSec->name + nLen + 1) = '\0';
    }

    /* parse flags if any */
    if (*pName == '.') {
	pName++;

	while (*pName) {
	    switch (*pName) {
	    case 'M':
		pSec->bSecFlags |= SECTION_DEF_FLAG_MPONLY;
		break;
	    case 'S':
		pSec->bSecFlags |= SECTION_DEF_FLAG_SPONLY;
		break;
	    }
	}
    }

    return 0;
}

/*
 * Method:	sortSecAddr
 *
 * Description:	do a simple bubble sort on the section list to get load
 *		addresses in accending order.  This makes it much easier
 *		for building images.
 *
 * Parameters:	obj	- elf read object
 *
 * Returns:	nothing
 */
static void
sortSecAddr(elfFileData_t *pObj)
{
    elfSection_t **pSecList, **pTestSecList;
    elfSection_t *pTemp;
    int i, j;

    for (i = pObj->numSections - 1, pSecList = pObj->pSecList; i > 0; i--, pSecList++) {
	for (j = i, pTestSecList = pSecList+1; j > 0; j--, pTestSecList++) {
	    if (((int)(*pSecList)->addr) > ((int)(*pTestSecList)->addr)) {
		pTemp = *pTestSecList;
		*pTestSecList = *pSecList;
		*pSecList = pTemp;
	    }
	}
    }
}
