/*
 * COPYRIGHT NOTICE
 * Copyright (c) Alteon Networks, Inc. 1998
 * All rights reserved
 * 
 */
/*
 * HISTORY
 * $Log: nextToken.c,v $
 * Revision 1.1.2.1  1998/06/19  07:55:47  tibor
 * 	new pasing algs
 * 	[1998/06/19  07:45:42  tibor]
 *
 * $EndLog$
 */

#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include "nextToken.h"

/*
 *
 * Function:	nextToken
 *
 * Purpose:	parse the input line and get the next syntactic element.
 *		An element is a C symbol, special character or character
 *		sequence.  a preprocessor system is treated as a symbol.
 *
 * Parms:	line	- line to parse or NULL
 *		token	- pointer to token string storage
 *		last	- if line is NULL, use starting point
 *
 * Returns:	1 	- token found
 *		0 	- no token
 *		-1 	- error
 *
 */
int
nextToken(char *line, char *token, char **last)
{
    int nLoc = 0;
    int nState = 0;
    int retcode = -1;
    char *p;
    
    /* line always has the start of parse location */
    if (line == NULL) {
	line = *last;
    } 

    /* initial the returned token is null */
    *token = '\0';

    /* skip to first tokenizable character */
    line += strspn(line, " \t\n");

    p = line;

    while (nState >= 0 && nState != 99) {
	switch (nState) {
	case 0:
	    /* if first char is a  # then treat as symbol */
	    if (*p == '#') {
		p++;
	    }

	    /* now do initial vectoring to other states */
	    if (*p == '_' || isalnum(*p)) {
		/* go to symbol eval */
		nState = 1;
	    } else if ((nLoc = strspn(p, "!%^&*-+=~|.<>/?"))) {
		/* go to operator eval */
		nState = 2;
	    } else if (strspn(p, "()[]{},;:\"'")) { 
		/* go to separator eval */
		nState = 3;
	    } else if (!*p) {
		/* blank line */
		nState = 99;
	    } else {
		nState = -1;
	    }
	    break;
	case 1:
	    /* symbol eval */
	    if (*p != '_' && !isalnum(*p)) {
		nState = 99;
	    } else {
		p++;
	    }
	    break;
	case 2:
	    /* operator type vectoring, simple operator step */
	    if (*p == '~' || *p == '.' || *p == '?' || nLoc == 1) {
		/* single char operator or operator delimited by non-oper */
		p++;
		nState = 99;
	    } else if (*(p+1) == '=') {
		/* easy two character token */
		p+=2;
		nState = 99;
	    } else if (*p == '!' || *p == '%' || *p == '^') {
		/* single char operator, next char is a new operator */
		p++;
		nState = 99;
	    } else if ((*p == '/' && *(p+1) == '*')
		       || (*p == '*' && *(p+1) == '/')
		       || (*p == '-' && *(p+1) == '>')) {
		/* comment set or -> */
		p+=2;
		nState = 99;
	    } else if (*p == *(p+1) && (*p != '/') && (*p != '*')) {
		/* operators <<,>>,++,--, ||, &&, >>=, <<= */
		if ((nLoc != 2)
		    && ((*p == '<' || *p == '>') && *(p+2) == '=')) {
		    /* operators <<=, >>= */
		    p++;
		}
		p+=2;
	    } else {
		/* some single character operator */
		p++;
		nState = 99;
	    }
	    break;
	case 3:
	    /* separator charater, only one character token */
	    p++;
	    nState = 99;
	    break;
	default:
	    (void) fprintf(stderr, "nextToken() error 1\n");
	    nState = -1;
	    break;
	}
    }

    nLoc = p - line;
    if (nState == 99) {
	if (nLoc > 0) {
	    strncpy (token, line, nLoc);
	    token[nLoc] = '\0';
	    retcode = 1;
	} else {
	    retcode = 0;
	}
    }
    
    *last = p;
    return retcode;
}

/*
 *
 * Function:	parseValue
 *
 * Purpose:	parse the input line at current strtok() location for
 *		num <oper> num [ <oper> num ...].  Return value of parse.
 *
 * Parms:	line 
 *		linenum - current line number for error callout
 *
 * Returns:	parsed value or zero for error. 
 *
 */
unsigned long
parseValue(char *line, int lineNum)
{
    char *pLast = line;
    char sToken[80];
    char *pToken = sToken;
    char *pEnd;
    unsigned long nValue = 0;
    int nState = 0;
    int nOperand = 0;
    unsigned long nNum1 = 0, nNum2;
    char oper = '\0';
    int base;

    while (nState >= 0 && nextToken(NULL, pToken, &pLast) == 1) {
	switch (nState) {
	case 0:
	    /* first number parse */
	    base = 10;
	    if (*pToken == '0' && toupper(*(pToken+1)) == 'X') {
		base = 16;
		pToken += 2;
	    }
	    
	    nNum1 = strtoul(pToken, &pEnd, base);
	    nOperand++;
	    if (*pEnd == '\0') {
		nState = 1;
	    } else {
		(void) fprintf(stderr, "Invalid %d operand '%s' on Line %d\n",
			       nOperand, sToken, lineNum);
		nState = -1;
	    }
	    break;
	case 1:
	case 3:
	    /* get operator */
	    if (strlen(pToken) == 1 && strpbrk(pToken, "+-*/")) {
		oper = *pToken;
		nState = 2;
	    } else {
		(void) fprintf(stderr, "Invalid operator on Line %d\n", lineNum);
		nState = -1;
	    }
	    break;
	case 2:
	    /* second number, compute result */
	    /* first number parse */
	    base = 10;
	    if (*pToken == '0' && toupper(*(pToken+1)) == 'X') {
		base = 16;
		pToken += 2;
	    }
	    
	    nNum2 = strtoul(pToken, &pEnd, base);
	    if (*pEnd == '\0') {
		nState = 3;

		switch (oper) {
		case '+':
		    nNum1 += nNum2;
		    break;
		case '-':
		    nNum1 -= nNum2;
		    break;
		case '*':
		    nNum1 *= nNum2;
		    break;
		case '/':
		    nNum1 /= nNum2;
		    break;
		default:
		    (void) fprintf(stderr, "parseValue() error 1\n");
		    nState = -1;
		    break;
		}

	    } else {
		(void) fprintf(stderr, "Invalid %d operand '%s' on Line %d\n",
			       nOperand, sToken, lineNum);
		nState = -1;
	    }
	    break;
	default:
	    (void) fprintf(stderr, "parseValue() error 2\n");
	    nState = -1;
	    break;
	}

	pToken = sToken;
    }

    if (nState == 1 || nState == 3) {
	nValue = nNum1;
    }
    
    return nValue;
}
